@@ -42,9 +42,11 @@
 #include "handler/RemotedHandler.h"
 #include "impl/StoredSession.h"
 #include "impl/StorageServiceSessionCache.h"
+#include "util/IPRange.h"
 #include "util/SPConstants.h"
 
 #include <algorithm>
+#define BOOST_BIND_GLOBAL_PLACEHOLDERS
 #include <boost/bind.hpp>
 #include <xmltooling/io/HTTPRequest.h>
 #include <xmltooling/io/HTTPResponse.h>
@@ -53,6 +55,7 @@
 #include <xmltooling/util/Threads.h>
 #include <xmltooling/util/URLEncoder.h>
 #include <xmltooling/util/XMLHelper.h>
+#include <xercesc/util/XMLStringTokenizer.hpp>
 #include <xercesc/util/XMLUniDefs.hpp>
 
 #ifndef SHIBSP_LITE
@@ -63,7 +66,6 @@
 # include <xmltooling/XMLToolingConfig.h>
 # include <xmltooling/util/ParserPool.h>
 # include <xmltooling/util/StorageService.h>
-# include <xercesc/util/XMLStringTokenizer.hpp>
 using namespace opensaml::saml2md;
 #else
 # include <xercesc/util/XMLDateTime.hpp>
@@ -117,6 +119,7 @@ SSCache::SSCache(const DOMElement* e, bool deprecationSupport)
     static const XMLCh outboundHeader[] =       UNICODE_LITERAL_14(o,u,t,b,o,u,n,d,H,e,a,d,e,r);
     static const XMLCh _StorageService[] =      UNICODE_LITERAL_14(S,t,o,r,a,g,e,S,e,r,v,i,c,e);
     static const XMLCh _StorageServiceLite[] =  UNICODE_LITERAL_18(S,t,o,r,a,g,e,S,e,r,v,i,c,e,L,i,t,e);
+    static const XMLCh _unreliableNetworks[] =  UNICODE_LITERAL_18(u,n,r,e,l,i,a,b,l,e,N,e,t,w,o,r,k,s);
 
     if (e && e->hasAttributeNS(nullptr, cacheTimeout)) {
         m_log.warn("DEPRECATED: cacheTimeout property is replaced by cacheAllowance (see documentation)");
@@ -189,6 +192,15 @@ SSCache::SSCache(const DOMElement* e, bool deprecationSupport)
     }
 #endif
 
+    const XMLCh* unreliableNetworks = e ? e->getAttributeNS(nullptr, _unreliableNetworks) : nullptr;
+    if (unreliableNetworks && *unreliableNetworks) {
+        XMLStringTokenizer toks(unreliableNetworks);
+        while (toks.hasMoreTokens()) {
+            auto_ptr_char tok(toks.nextToken());
+            m_unreliableNetworks.push_back(IPRange::parseCIDRBlock(tok.get()));
+        }
+    }
+
     ListenerService* listener=conf.getServiceProvider()->getListenerService(false);
     if (inproc) {
         if (!conf.isEnabled(SPConfig::OutOfProcess) && !listener)
@@ -272,11 +284,26 @@ string SSCache::active(const Application& app, const HTTPRequest& request)
         if (!session_id.empty())
             return session_id;
     }
-    pair<string, const char*> shib_cookie = app.getCookieNameProps("_shibsession_");
-    const char* session_id = request.getCookie(shib_cookie.first.c_str());
+
+    const char* session_id = request.getCookie(app.getCookieName("_shibsession_").c_str());
     return (session_id ? session_id : "");
 }
 
+bool SSCache::compareAddresses(const char* client_addr, const char* session_addr) const
+{
+    if (XMLString::equals(client_addr, session_addr)) {
+        return true;
+    }
+
+    for (vector<IPRange>::const_iterator i = m_unreliableNetworks.begin(); i != m_unreliableNetworks.end(); ++i) {
+        if (i->contains(client_addr) && i->contains(session_addr)) {
+            return true;
+        }
+    }
+
+    return false;
+}
+
 #ifndef SHIBSP_LITE
 
 void SSCache::test()
@@ -529,32 +556,24 @@ void SSCache::insert(
         httpResponse.setResponseHeader(m_outboundHeader.c_str(), key.get());
 
     time_t cookieLifetime = 0;
-    pair<string,const char*> shib_cookie = app.getCookieNameProps("_shibsession_", &cookieLifetime);
-    string k(key.get());
-    k += shib_cookie.second;
-
-    if (cookieLifetime > 0) {
-        cookieLifetime += now;
-#ifndef HAVE_GMTIME_R
-        ptime=gmtime(&cookieLifetime);
-#else
-        ptime=gmtime_r(&cookieLifetime,&res);
-#endif
-        char cookietimebuf[64];
-        strftime(cookietimebuf,64,"; expires=%a, %d %b %Y %H:%M:%S GMT",ptime);
-        k += cookietimebuf;
-    }
-
-    httpResponse.setCookie(shib_cookie.first.c_str(), k.c_str());
+    string shib_cookie = app.getCookieName("_shibsession_", &cookieLifetime);
+    HTTPResponse::samesite_t sameSitePolicy = getSameSitePolicy(app);
+    httpResponse.setCookie(shib_cookie.c_str(), key.get(), cookieLifetime, sameSitePolicy);
     sessionID = key.get();
 
     // See if we need to persist the session data itself to a cookie for cross-node recovery.
     if (!m_persistedAttributeIds.empty()) {
-        persist(app, httpResponse, obj, expires);
+        persist(app, httpResponse, obj, expires, sameSitePolicy);
     }
 }
 
-void SSCache::persist(const Application& app, HTTPResponse& httpResponse, DDF& session, time_t expires) const
+void SSCache::persist(
+    const Application& app,
+    HTTPResponse& httpResponse,
+    DDF& session,
+    time_t expires,
+    HTTPResponse::samesite_t sameSitePolicy
+    ) const
 {
 #ifdef _DEBUG
     xmltooling::NDC ndc("persist");
@@ -593,21 +612,8 @@ void SSCache::persist(const Application& app, HTTPResponse& httpResponse, DDF& s
         sealed = XMLToolingConfig::getConfig().getURLEncoder()->encode(sealed.c_str());
 
         time_t cookieLifetime;
-        pair<string,const char*> shib_cookie = app.getCookieNameProps("_shibsealed_", &cookieLifetime);
-        sealed += shib_cookie.second;
-        if (cookieLifetime > 0) {
-            cookieLifetime += time(nullptr);
-#ifndef HAVE_GMTIME_R
-            struct tm* ptime = gmtime(&cookieLifetime);
-#else
-            struct tm res;
-            struct tm* ptime = gmtime_r(&cookieLifetime, &res);
-#endif
-            char cookietimebuf[64];
-            strftime(cookietimebuf, 64, "; expires=%a, %d %b %Y %H:%M:%S GMT", ptime);
-            sealed += cookietimebuf;
-        }
-        httpResponse.setCookie(shib_cookie.first.c_str(), sealed.c_str());
+        string shib_cookie = app.getCookieName("_shibsealed_", &cookieLifetime);
+        httpResponse.setCookie(shib_cookie.c_str(), sealed.c_str(), cookieLifetime, sameSitePolicy);
     }
     catch (const std::exception& e) {
         m_log.error("failed to wrap session (%s) with DataSealer: %s", session.name(), e.what());
@@ -870,6 +876,26 @@ LogoutEvent* SSCache::newLogoutEvent(const Application& app) const
 
 #endif
 
+HTTPResponse::samesite_t SSCache::getSameSitePolicy(const Application& app) const
+{
+    const PropertySet* props = app.getPropertySet("Sessions");
+    if (props) {
+        pair<bool,const char*> sameSiteSession = props->getString("sameSiteSession");
+        if (sameSiteSession.first) {
+            if (!strcmp(sameSiteSession.second, "None")) {
+                return HTTPResponse::SAMESITE_NONE;
+            }
+            else if (!strcmp(sameSiteSession.second, "Lax")) {
+                return HTTPResponse::SAMESITE_LAX;
+            }
+            else if (!strcmp(sameSiteSession.second, "Strict")) {
+                return HTTPResponse::SAMESITE_STRICT;
+            }
+        }
+    }
+    return HTTPResponse::SAMESITE_ABSENT;
+}
+
 Session* SSCache::_find(const Application& app, const char* key, const char* recovery, const char* client_addr, time_t* timeout)
 {
 #ifdef _DEBUG
@@ -1064,8 +1090,8 @@ Session* SSCache::find(const Application& app, HTTPRequest& request, const char*
     if (id.empty())
         return nullptr;
 
-    pair<string, const char*> shib_cookie = app.getCookieNameProps("_shibsealed_");
-    const char* c = request.getCookie(shib_cookie.first.c_str());
+    HTTPResponse::samesite_t sameSitePolicy = getSameSitePolicy(app);
+    const char* c = request.getCookie(app.getCookieName("_shibsealed_").c_str());
 
     try {
         Session* session = _find(app, id.c_str(), c, client_addr, timeout);
@@ -1076,12 +1102,8 @@ Session* SSCache::find(const Application& app, HTTPRequest& request, const char*
         if (response) {
             if (!m_outboundHeader.empty())
                 response->setResponseHeader(m_outboundHeader.c_str(), nullptr);
-            pair<string,const char*> shib_cookie = app.getCookieNameProps("_shibsession_");
-            string exp(shib_cookie.second);
-            exp += "; expires=Mon, 01 Jan 2001 00:00:00 GMT";
-            response->setCookie(shib_cookie.first.c_str(), exp.c_str());
-            shib_cookie = app.getCookieNameProps("_shibsealed_");
-            response->setCookie(shib_cookie.first.c_str(), exp.c_str());
+            response->setCookie(app.getCookieName("_shibsession_").c_str(), nullptr, 0, sameSitePolicy);
+            response->setCookie(app.getCookieName("_shibsealed_").c_str(), nullptr, 0, sameSitePolicy);
         }
     }
     catch (const std::exception&) {
@@ -1089,12 +1111,8 @@ Session* SSCache::find(const Application& app, HTTPRequest& request, const char*
         if (response) {
             if (!m_outboundHeader.empty())
                 response->setResponseHeader(m_outboundHeader.c_str(), nullptr);
-            pair<string,const char*> shib_cookie = app.getCookieNameProps("_shibsession_");
-            string exp(shib_cookie.second);
-            exp += "; expires=Mon, 01 Jan 2001 00:00:00 GMT";
-            response->setCookie(shib_cookie.first.c_str(), exp.c_str());
-            shib_cookie = app.getCookieNameProps("_shibsealed_");
-            response->setCookie(shib_cookie.first.c_str(), exp.c_str());
+            response->setCookie(app.getCookieName("_shibsession_").c_str(), nullptr, 0, sameSitePolicy);
+            response->setCookie(app.getCookieName("_shibsealed_").c_str(), nullptr, 0, sameSitePolicy);
         }
         throw;
     }
@@ -1228,12 +1246,12 @@ void SSCache::remove(const Application& app, const HTTPRequest& request, HTTPRes
     xmltooling::NDC ndc("remove");
 #endif
     string session_id;
-    pair<string,const char*> shib_cookie = app.getCookieNameProps("_shibsession_");
+    string shib_cookie = app.getCookieName("_shibsession_");
 
     if (!m_inboundHeader.empty())
         session_id = request.getHeader(m_inboundHeader.c_str());
     if (session_id.empty()) {
-        const char* c = request.getCookie(shib_cookie.first.c_str());
+        const char* c = request.getCookie(shib_cookie.c_str());
         if (c && *c)
             session_id = c;
     }
@@ -1242,12 +1260,9 @@ void SSCache::remove(const Application& app, const HTTPRequest& request, HTTPRes
         if (response) {
             if (!m_outboundHeader.empty())
                 response->setResponseHeader(m_outboundHeader.c_str(), nullptr);
-            string exp(shib_cookie.second);
-            exp += "; expires=Mon, 01 Jan 2001 00:00:00 GMT";
-            response->setCookie(shib_cookie.first.c_str(), exp.c_str());
-
-            shib_cookie = app.getCookieNameProps("_shibsealed_");
-            response->setCookie(shib_cookie.first.c_str(), exp.c_str());
+            HTTPResponse::samesite_t sameSitePolicy = getSameSitePolicy(app);
+            response->setCookie(shib_cookie.c_str(), nullptr, 0, sameSitePolicy);
+            response->setCookie(app.getCookieName("_shibsealed_").c_str(), nullptr, 0, sameSitePolicy);
         }
         remove(app, session_id.c_str(), revocationExp);
     }
